@@ -365,7 +365,6 @@ static void print_one_uuid(struct btrfs_fs_devices *fs_devices,
 	if (devs_found < total) {
 		pr_verbose(LOG_DEFAULT, "\t*** Some devices missing\n");
 	}
-	pr_verbose(LOG_DEFAULT, "\n");
 }
 
 /* adds up all the used spaces as reported by the space info ioctl
@@ -431,7 +430,6 @@ static int print_one_fs(struct btrfs_ioctl_fs_info_args *fs_info,
 		free(canonical_path);
 	}
 
-	pr_verbose(LOG_DEFAULT, "\n");
 	return 0;
 }
 
@@ -481,6 +479,10 @@ static int btrfs_scan_kernel(void *search, unsigned unit_mode)
 
 		fd = open(mnt->mnt_dir, O_RDONLY);
 		if ((fd != -1) && !get_df(fd, &space_info_arg)) {
+			/* Put space between filesystem entries for readability. */
+			if (found != 0)
+				pr_verbose(LOG_DEFAULT, "\n");
+
 			print_one_fs(&fs_info_arg, dev_info_arg,
 				     space_info_arg, label, unit_mode);
 			free(space_info_arg);
@@ -757,6 +759,7 @@ static int cmd_filesystem_show(const struct cmd_struct *cmd,
 	char uuid_buf[BTRFS_UUID_UNPARSED_SIZE];
 	unsigned unit_mode;
 	int found = 0;
+	bool needs_newline = false;
 
 	unit_mode = get_unit_mode_from_arg(&argc, argv, 0);
 
@@ -845,6 +848,12 @@ static int cmd_filesystem_show(const struct cmd_struct *cmd,
 		goto out;
 	}
 
+	/*
+	 * The above call will return 0 if it found anything, in those cases we
+	 * need an extra newline below.
+	 */
+	needs_newline = !ret;
+
 	/* shows mounted only */
 	if (where == BTRFS_SCAN_MOUNTED)
 		goto out;
@@ -882,8 +891,14 @@ devs_only:
 		goto out;
 	}
 
-	list_for_each_entry(fs_devices, &all_uuids, fs_list)
+	list_for_each_entry(fs_devices, &all_uuids, fs_list) {
+		/* Put space between filesystem entries for readability. */
+		if (needs_newline)
+			pr_verbose(LOG_DEFAULT, "\n");
+
 		print_one_uuid(fs_devices, unit_mode);
+		needs_newline = true;
+	}
 
 	if (search && !found) {
 		error("not a valid btrfs filesystem: %s", search);
@@ -919,7 +934,7 @@ static int cmd_filesystem_sync(const struct cmd_struct *cmd,
 	if (check_argc_exact(argc - optind, 1))
 		return 1;
 
-	err = btrfs_util_sync(argv[optind]);
+	err = btrfs_util_fs_sync(argv[optind]);
 	if (err) {
 		error_btrfs_util(err);
 		return 1;
@@ -947,6 +962,7 @@ static const char * const cmd_filesystem_defrag_usage[] = {
 	"",
 	OPTLINE("-r", "defragment files recursively"),
 	OPTLINE("-c[zlib,lzo,zstd]", "compress the file while defragmenting, optional parameter (no space in between)"),
+	OPTLINE("-L|--level level", "use given compression level if enabled (zlib: 1..9, zstd: -15..15, and 0 selects the default level)"),
 	OPTLINE("-f", "flush data to disk immediately after defragmenting"),
 	OPTLINE("-s start", "defragment only from byte onward"),
 	OPTLINE("-l len", "defragment only up to len bytes"),
@@ -1051,6 +1067,7 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 	bool recursive = false;
 	int ret = 0;
 	int compress_type = BTRFS_COMPRESS_NONE;
+	int compress_level = 0;
 
 	/*
 	 * Kernel 4.19+ supports defragmention of files open read-only,
@@ -1080,18 +1097,18 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 	if (bconf.verbose != BTRFS_BCONF_UNSET)
 		bconf.verbose++;
 
-	defrag_global_errors = 0;
 	defrag_global_errors = 0;
 	optind = 0;
 	while(1) {
 		enum { GETOPT_VAL_STEP = GETOPT_VAL_FIRST };
 		static const struct option long_options[] = {
+			{ "level", required_argument, NULL, 'L' },
 			{ "step", required_argument, NULL, GETOPT_VAL_STEP },
 			{ NULL, 0, NULL, 0 }
 		};
 		int c;
 
-		c = getopt_long(argc, argv, "vrc::fs:l:t:", long_options, NULL);
+		c = getopt_long(argc, argv, "vrc::L:fs:l:t:", long_options, NULL);
 		if (c < 0)
 			break;
 
@@ -1101,6 +1118,18 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 			if (optarg)
 				compress_type = parse_compress_type_arg(optarg);
 			break;
+		case 'L':
+			/*
+			 * Do not enforce any limits here, kernel will do itself
+			 * based on what's supported by the running version.
+			 * Just clip to the s8 type of the API.
+			 */
+			compress_level = atoi(optarg);
+			if (compress_level < -128)
+				compress_level = -128;
+			else if (compress_level > 127)
+				compress_level = 127;
+			break;
 		case 'f':
 			flush = true;
 			break;
@@ -1150,7 +1179,12 @@ static int cmd_filesystem_defrag(const struct cmd_struct *cmd,
 	defrag_global_range.extent_thresh = (u32)thresh;
 	if (compress_type) {
 		defrag_global_range.flags |= BTRFS_DEFRAG_RANGE_COMPRESS;
-		defrag_global_range.compress_type = compress_type;
+		if (compress_level) {
+			defrag_global_range.flags |= BTRFS_DEFRAG_RANGE_COMPRESS_LEVEL;
+			defrag_global_range.compress.type = compress_type;
+			defrag_global_range.compress.level= compress_level;
+		} else
+			defrag_global_range.compress_type = compress_type;
 	}
 	if (flush)
 		defrag_global_range.flags |= BTRFS_DEFRAG_RANGE_START_IO;
@@ -1254,7 +1288,8 @@ static const char * const cmd_filesystem_resize_usage[] = {
 	NULL
 };
 
-static int check_resize_args(const char *amount, const char *path, u64 *devid_ret) {
+static int check_resize_args(const char *amount, const char *path, u64 *devid_ret)
+{
 	struct btrfs_ioctl_fs_info_args fi_args;
 	struct btrfs_ioctl_dev_info_args *di_args = NULL;
 	int ret, i, dev_idx = -1;
@@ -1387,6 +1422,10 @@ static int check_resize_args(const char *amount, const char *path, u64 *devid_re
 		}
 		new_size = round_down(new_size, fi_args.sectorsize);
 		res_str = pretty_size_mode(new_size, UNITS_DEFAULT);
+
+		if (new_size < 256 * SZ_1M)
+   warning("the new size %lld (%s) is < 256MiB, this may be rejected by kernel",
+			new_size, pretty_size_mode(new_size, UNITS_DEFAULT));
 	}
 
 	pr_verbose(LOG_DEFAULT, "Resize device id %lld (%s) from %s to %s\n", devid,
