@@ -24,6 +24,15 @@
 #include "kerncompat.h"
 #include <sys/types.h>
 #include <stdbool.h>
+#include <limits.h>
+#include "kernel-lib/list.h"
+#include "kernel-shared/compression.h"
+
+#define ZLIB_BTRFS_DEFAULT_LEVEL		3
+#define ZLIB_BTRFS_MAX_LEVEL			9
+
+#define ZSTD_BTRFS_DEFAULT_LEVEL		3
+#define ZSTD_BTRFS_MAX_LEVEL			15
 
 struct btrfs_fs_info;
 struct btrfs_root;
@@ -37,7 +46,9 @@ struct rootdir_subvol {
 };
 
 int btrfs_mkfs_fill_dir(struct btrfs_trans_handle *trans, const char *source_dir,
-			struct btrfs_root *root, struct list_head *subvols);
+			struct btrfs_root *root, struct list_head *subvols,
+			enum btrfs_compression_type compression,
+			unsigned int compression_level);
 u64 btrfs_mkfs_size_dir(const char *dir_name, u32 sectorsize, u64 min_dev_size,
 			u64 meta_profile, u64 data_profile);
 int btrfs_mkfs_shrink_fs(struct btrfs_fs_info *fs_info, u64 *new_size_ret,
